"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.validateLoaderCompatibility = exports.LoaderSupportRequirements = exports.RuntimeCompatDetails = void 0;
const client_utils_1 = require("@fluid-internal/client-utils");
const internal_1 = require("@fluidframework/telemetry-utils/internal");
const packageVersion_js_1 = require("./packageVersion.js");
/**
 * Runtime's compatibility details that is exposed to the Loader layer.
 */
exports.RuntimeCompatDetails = {
    /**
     * The package version of the Runtime layer.
     */
    pkgVersion: packageVersion_js_1.pkgVersion,
    /**
     * The current generation of the Runtime layer.
     */
    generation: 1,
    /**
     * The features supported by the Runtime layer across the Runtime / Loader boundary.
     */
    supportedFeatures: new Set(),
};
/**
 * The requirements that the Loader layer must meet to be compatible with this Runtime.
 */
exports.LoaderSupportRequirements = {
    /**
     * Minimum generation that Loader must be at to be compatible with Runtime.
     */
    minSupportedGeneration: 0,
    /**
     * The features that the Loader must support to be compatible with Runtime. Note that 0 is used here for
     * Loader layers before the introduction of the layer compatibility enforcement.
     */
    requiredFeatures: [],
};
/**
 * Validates that the Loader layer is compatible with this Runtime.
 */
function validateLoaderCompatibility(maybeLoaderCompatDetails, disposeFn) {
    const layerCheckResult = (0, client_utils_1.checkLayerCompatibility)(exports.LoaderSupportRequirements, maybeLoaderCompatDetails);
    if (!layerCheckResult.isCompatible) {
        const error = new internal_1.UsageError("Runtime is not compatible with Loader", {
            errorDetails: JSON.stringify({
                runtimeVersion: exports.RuntimeCompatDetails.pkgVersion,
                loaderVersion: maybeLoaderCompatDetails?.pkgVersion,
                runtimeGeneration: exports.RuntimeCompatDetails.generation,
                loaderGeneration: maybeLoaderCompatDetails?.generation,
                minSupportedGeneration: exports.LoaderSupportRequirements.minSupportedGeneration,
                isGenerationCompatible: layerCheckResult.isGenerationCompatible,
                unsupportedFeatures: layerCheckResult.unsupportedFeatures,
            }),
        });
        disposeFn(error);
        throw error;
    }
}
exports.validateLoaderCompatibility = validateLoaderCompatibility;
//# sourceMappingURL=layerCompatState.js.map