"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.Outbox = exports.getLongStack = exports.serializeOpContents = void 0;
const internal_1 = require("@fluidframework/core-utils/internal");
const internal_2 = require("@fluidframework/telemetry-utils/internal");
const batchManager_js_1 = require("./batchManager.js");
/**
 * Before submitting an op to the Outbox, its contents must be serialized using this function.
 * @remarks - The deserialization on process happens via the function {@link ensureContentsDeserialized}.
 */
function serializeOpContents(contents) {
    return JSON.stringify(contents);
}
exports.serializeOpContents = serializeOpContents;
/**
 * Temporarily increase the stack limit while executing the provided action.
 * If a negative value is provided for `length`, no stack frames will be collected.
 * If Infinity is provided, all frames will be collected.
 *
 * ADO:4663 - add this to the common packages.
 *
 * @param action - action which returns an error
 * @param length - number of stack frames to collect, 50 if unspecified.
 * @returns the result of the action provided
 */
function getLongStack(action, length = 50) {
    // TODO: better typing here
    // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-assignment
    const errorObj = Error;
    if (
    /* eslint-disable @typescript-eslint/prefer-nullish-coalescing */
    // ?? is not logically equivalent when the first clause returns false.
    (Object.getOwnPropertyDescriptor(errorObj, "stackTraceLimit") ||
        Object.getOwnPropertyDescriptor(Object.getPrototypeOf(errorObj), "stackTraceLimit"))?.writable !== true
    /* eslint-enable @typescript-eslint/prefer-nullish-coalescing */
    ) {
        return action();
    }
    // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access, @typescript-eslint/no-unsafe-assignment
    const originalStackTraceLimit = errorObj.stackTraceLimit;
    try {
        // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
        errorObj.stackTraceLimit = length;
        return action();
    }
    finally {
        // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access, @typescript-eslint/no-unsafe-assignment
        errorObj.stackTraceLimit = originalStackTraceLimit;
    }
}
exports.getLongStack = getLongStack;
/**
 * The Outbox collects messages submitted by the ContainerRuntime into a batch,
 * and then flushes the batch when requested.
 *
 * @remarks There are actually multiple independent batches (some are for a specific message type),
 * to support slight variation in semantics for each batch (e.g. support for rebasing or grouping).
 */
class Outbox {
    constructor(params) {
        this.params = params;
        this.batchRebasesToReport = 5;
        this.rebasing = false;
        /**
         * Track the number of ops which were detected to have a mismatched
         * reference sequence number, in order to self-throttle the telemetry events.
         *
         * This should be removed as part of ADO:2322
         */
        this.maxMismatchedOpsToReport = 3;
        this.mismatchedOpsReported = 0;
        this.logger = (0, internal_2.createChildLogger)({ logger: params.logger, namespace: "Outbox" });
        const isCompressionEnabled = this.params.config.compressionOptions.minimumBatchSizeInBytes !==
            Number.POSITIVE_INFINITY;
        // We need to allow infinite size batches if we enable compression
        const hardLimit = isCompressionEnabled
            ? Number.POSITIVE_INFINITY
            : this.params.config.maxBatchSizeInBytes;
        this.mainBatch = new batchManager_js_1.BatchManager({ hardLimit, canRebase: true });
        this.blobAttachBatch = new batchManager_js_1.BatchManager({ hardLimit, canRebase: true });
        this.idAllocationBatch = new batchManager_js_1.BatchManager({
            hardLimit,
            canRebase: false,
            ignoreBatchId: true,
        });
    }
    get messageCount() {
        return this.mainBatch.length + this.blobAttachBatch.length + this.idAllocationBatch.length;
    }
    get mainBatchMessageCount() {
        return this.mainBatch.length;
    }
    get blobAttachBatchMessageCount() {
        return this.blobAttachBatch.length;
    }
    get idAllocationBatchMessageCount() {
        return this.idAllocationBatch.length;
    }
    get isEmpty() {
        return this.messageCount === 0;
    }
    /**
     * Detect whether batching has been interrupted by an incoming message being processed. In this case,
     * we will flush the accumulated messages to account for that and create a new batch with the new
     * message as the first message.
     *
     * @remarks - To detect batch interruption, we compare both the reference sequence number
     * (i.e. last message processed by DeltaManager) and the client sequence number of the
     * last message processed by the ContainerRuntime. In the absence of op reentrancy, this
     * pair will remain stable during a single JS turn during which the batch is being built up.
     */
    maybeFlushPartialBatch() {
        const mainBatchSeqNums = this.mainBatch.sequenceNumbers;
        const blobAttachSeqNums = this.blobAttachBatch.sequenceNumbers;
        const idAllocSeqNums = this.idAllocationBatch.sequenceNumbers;
        (0, internal_1.assert)(this.params.config.disablePartialFlush ||
            ((0, batchManager_js_1.sequenceNumbersMatch)(mainBatchSeqNums, blobAttachSeqNums) &&
                (0, batchManager_js_1.sequenceNumbersMatch)(mainBatchSeqNums, idAllocSeqNums)), 0x58d /* Reference sequence numbers from both batches must be in sync */);
        const currentSequenceNumbers = this.params.getCurrentSequenceNumbers();
        if ((0, batchManager_js_1.sequenceNumbersMatch)(mainBatchSeqNums, currentSequenceNumbers) &&
            (0, batchManager_js_1.sequenceNumbersMatch)(blobAttachSeqNums, currentSequenceNumbers) &&
            (0, batchManager_js_1.sequenceNumbersMatch)(idAllocSeqNums, currentSequenceNumbers)) {
            // The reference sequence numbers are stable, there is nothing to do
            return;
        }
        if (++this.mismatchedOpsReported <= this.maxMismatchedOpsToReport) {
            this.logger.sendTelemetryEvent({
                category: this.params.config.disablePartialFlush ? "error" : "generic",
                eventName: "ReferenceSequenceNumberMismatch",
                mainReferenceSequenceNumber: mainBatchSeqNums.referenceSequenceNumber,
                mainClientSequenceNumber: mainBatchSeqNums.clientSequenceNumber,
                blobAttachReferenceSequenceNumber: blobAttachSeqNums.referenceSequenceNumber,
                blobAttachClientSequenceNumber: blobAttachSeqNums.clientSequenceNumber,
                currentReferenceSequenceNumber: currentSequenceNumbers.referenceSequenceNumber,
                currentClientSequenceNumber: currentSequenceNumbers.clientSequenceNumber,
            }, getLongStack(() => new internal_2.UsageError("Submission of an out of order message")));
        }
        if (!this.params.config.disablePartialFlush) {
            this.flushAll();
        }
    }
    submit(message) {
        this.maybeFlushPartialBatch();
        this.addMessageToBatchManager(this.mainBatch, message);
    }
    submitBlobAttach(message) {
        this.maybeFlushPartialBatch();
        this.addMessageToBatchManager(this.blobAttachBatch, message);
    }
    submitIdAllocation(message) {
        this.maybeFlushPartialBatch();
        this.addMessageToBatchManager(this.idAllocationBatch, message);
    }
    addMessageToBatchManager(batchManager, message) {
        if (!batchManager.push(message, this.isContextReentrant(), this.params.getCurrentSequenceNumbers().clientSequenceNumber)) {
            throw new internal_2.GenericError("BatchTooLarge", /* error */ undefined, {
                opSize: message.contents?.length ?? 0,
                batchSize: batchManager.contentSizeInBytes,
                count: batchManager.length,
                limit: batchManager.options.hardLimit,
            });
        }
    }
    /**
     * Flush all the batches to the ordering service.
     * This method is expected to be called at the end of a batch.
     * @param resubmittingBatchId - If defined, indicates this is a resubmission of a batch
     * with the given Batch ID, which must be preserved
     */
    flush(resubmittingBatchId) {
        if (this.isContextReentrant()) {
            const error = new internal_2.UsageError("Flushing is not supported inside DDS event handlers");
            this.params.closeContainer(error);
            throw error;
        }
        this.flushAll(resubmittingBatchId);
    }
    flushAll(resubmittingBatchId) {
        // If we're resubmitting and all batches are empty, we need to flush an empty batch.
        // Note that we currently resubmit one batch at a time, so on resubmit, 2 of the 3 batches will *always* be empty.
        // It's theoretically possible that we don't *need* to resubmit this empty batch, and in those cases, it'll safely be ignored
        // by the rest of the system, including remote clients.
        // In some cases we *must* resubmit the empty batch (to match up with a non-empty version tracked locally by a container fork), so we do it always.
        const allBatchesEmpty = this.idAllocationBatch.empty && this.blobAttachBatch.empty && this.mainBatch.empty;
        if (resubmittingBatchId && allBatchesEmpty) {
            this.flushEmptyBatch(resubmittingBatchId);
            return;
        }
        // Don't use resubmittingBatchId for idAllocationBatch.
        // ID Allocation messages are not directly resubmitted so we don't want to reuse the batch ID.
        this.flushInternal(this.idAllocationBatch);
        this.flushInternal(this.blobAttachBatch, true /* disableGroupedBatching */, resubmittingBatchId);
        this.flushInternal(this.mainBatch, false /* disableGroupedBatching */, resubmittingBatchId);
    }
    flushEmptyBatch(resubmittingBatchId) {
        const referenceSequenceNumber = this.params.getCurrentSequenceNumbers().referenceSequenceNumber;
        (0, internal_1.assert)(referenceSequenceNumber !== undefined, 0xa01 /* reference sequence number should be defined */);
        const emptyGroupedBatch = this.params.groupingManager.createEmptyGroupedBatch(resubmittingBatchId, referenceSequenceNumber);
        let clientSequenceNumber;
        if (this.params.shouldSend()) {
            clientSequenceNumber = this.sendBatch(emptyGroupedBatch);
        }
        this.params.pendingStateManager.onFlushBatch(emptyGroupedBatch.messages, // This is the single empty Grouped Batch message
        clientSequenceNumber);
        return;
    }
    flushInternal(batchManager, disableGroupedBatching = false, resubmittingBatchId) {
        if (batchManager.empty) {
            return;
        }
        const rawBatch = batchManager.popBatch(resubmittingBatchId);
        const shouldGroup = !disableGroupedBatching && this.params.groupingManager.shouldGroup(rawBatch);
        if (batchManager.options.canRebase && rawBatch.hasReentrantOps === true && shouldGroup) {
            (0, internal_1.assert)(!this.rebasing, 0x6fa /* A rebased batch should never have reentrant ops */);
            // If a batch contains reentrant ops (ops created as a result from processing another op)
            // it needs to be rebased so that we can ensure consistent reference sequence numbers
            // and eventual consistency at the DDS level.
            this.rebase(rawBatch, batchManager);
            return;
        }
        let clientSequenceNumber;
        // Did we disconnect? (i.e. is shouldSend false?)
        // If so, do nothing, as pending state manager will resubmit it correctly on reconnect.
        // Because flush() is a task that executes async (on clean stack), we can get here in disconnected state.
        if (this.params.shouldSend()) {
            const processedBatch = disableGroupedBatching
                ? rawBatch
                : this.compressAndChunkBatch(shouldGroup ? this.params.groupingManager.groupBatch(rawBatch) : rawBatch);
            clientSequenceNumber = this.sendBatch(processedBatch);
            (0, internal_1.assert)(clientSequenceNumber === undefined || clientSequenceNumber >= 0, 0x9d2 /* unexpected negative clientSequenceNumber (empty batch should yield undefined) */);
        }
        this.params.pendingStateManager.onFlushBatch(rawBatch.messages, clientSequenceNumber, batchManager.options.ignoreBatchId);
    }
    /**
     * Rebases a batch. All the ops in the batch are resubmitted to the runtime and
     * they will end up back in the same batch manager they were flushed from and subsequently flushed.
     *
     * @param rawBatch - the batch to be rebased
     */
    rebase(rawBatch, batchManager) {
        (0, internal_1.assert)(!this.rebasing, 0x6fb /* Reentrancy */);
        (0, internal_1.assert)(batchManager.options.canRebase, 0x9a7 /* BatchManager does not support rebase */);
        this.rebasing = true;
        for (const message of rawBatch.messages) {
            this.params.reSubmit({
                // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
                content: message.contents,
                localOpMetadata: message.localOpMetadata,
                opMetadata: message.metadata,
            });
        }
        if (this.batchRebasesToReport > 0) {
            this.logger.sendTelemetryEvent({
                eventName: "BatchRebase",
                length: rawBatch.messages.length,
                referenceSequenceNumber: rawBatch.referenceSequenceNumber,
            }, new internal_2.UsageError("BatchRebase"));
            this.batchRebasesToReport--;
        }
        this.flushInternal(batchManager);
        this.rebasing = false;
    }
    isContextReentrant() {
        return this.params.opReentrancy() && !this.rebasing;
    }
    /**
     * As necessary and enabled, compresses and chunks the given batch.
     *
     * @remarks - If chunking happens, a side effect here is that 1 or more chunks are queued immediately for sending in next JS turn.
     *
     * @param batch - Raw or Grouped batch to consider for compression/chunking
     * @returns Either (A) the original batch, (B) a compressed batch (same length as original)
     * or (C) a batch containing the last chunk.
     */
    compressAndChunkBatch(batch) {
        if (batch.messages.length === 0 ||
            this.params.config.compressionOptions === undefined ||
            this.params.config.compressionOptions.minimumBatchSizeInBytes >
                batch.contentSizeInBytes ||
            this.params.submitBatchFn === undefined ||
            !this.params.groupingManager.groupedBatchingEnabled()) {
            // Nothing to do if the batch is empty or if compression is disabled or not supported, or if we don't need to compress
            return batch;
        }
        const compressedBatch = this.params.compressor.compressBatch(batch);
        if (this.params.splitter.isBatchChunkingEnabled) {
            return compressedBatch.contentSizeInBytes <= this.params.splitter.chunkSizeInBytes
                ? compressedBatch
                : this.params.splitter.splitFirstBatchMessage(compressedBatch);
        }
        if (compressedBatch.contentSizeInBytes >= this.params.config.maxBatchSizeInBytes) {
            throw new internal_2.GenericError("BatchTooLarge", /* error */ undefined, {
                batchSize: batch.contentSizeInBytes,
                compressedBatchSize: compressedBatch.contentSizeInBytes,
                count: compressedBatch.messages.length,
                limit: this.params.config.maxBatchSizeInBytes,
                chunkingEnabled: this.params.splitter.isBatchChunkingEnabled,
                compressionOptions: JSON.stringify(this.params.config.compressionOptions),
                socketSize: (0, batchManager_js_1.estimateSocketSize)(batch),
            });
        }
        return compressedBatch;
    }
    /**
     * Sends the batch object to the container context to be sent over the wire.
     *
     * @param batch - batch to be sent
     * @returns the clientSequenceNumber of the start of the batch, or undefined if nothing was sent
     */
    sendBatch(batch) {
        const length = batch.messages.length;
        if (length === 0) {
            return undefined; // Nothing submitted
        }
        const socketSize = (0, batchManager_js_1.estimateSocketSize)(batch);
        if (socketSize >= this.params.config.maxBatchSizeInBytes) {
            this.logger.sendPerformanceEvent({
                eventName: "LargeBatch",
                length: batch.messages.length,
                sizeInBytes: batch.contentSizeInBytes,
                socketSize,
            });
        }
        let clientSequenceNumber;
        if (this.params.submitBatchFn === undefined) {
            // Legacy path - supporting old loader versions. Can be removed only when LTS moves above
            // version that has support for batches (submitBatchFn)
            (0, internal_1.assert)(batch.messages[0].compression === undefined, 0x5a6 /* Compression should not have happened if the loader does not support it */);
            clientSequenceNumber = this.params.legacySendBatchFn(batch);
        }
        else {
            (0, internal_1.assert)(batch.referenceSequenceNumber !== undefined, 0x58e /* Batch must not be empty */);
            clientSequenceNumber = this.params.submitBatchFn(batch.messages.map((message) => ({
                contents: message.contents,
                metadata: message.metadata,
                compression: message.compression,
                referenceSequenceNumber: message.referenceSequenceNumber,
            })), batch.referenceSequenceNumber);
        }
        // Convert from clientSequenceNumber of last message in the batch to clientSequenceNumber of first message.
        clientSequenceNumber -= length - 1;
        (0, internal_1.assert)(clientSequenceNumber >= 0, 0x3d0 /* clientSequenceNumber can't be negative */);
        return clientSequenceNumber;
    }
    /**
     * Gets a checkpoint object per batch that facilitates iterating over the batch messages when rolling back.
     */
    getBatchCheckpoints() {
        // This variable is declared with a specific type so that we have a standard import of the IBatchCheckpoint type.
        // When the type is inferred, the generated .d.ts uses a dynamic import which doesn't resolve.
        const mainBatch = this.mainBatch.checkpoint();
        return {
            mainBatch,
            idAllocationBatch: this.idAllocationBatch.checkpoint(),
            blobAttachBatch: this.blobAttachBatch.checkpoint(),
        };
    }
}
exports.Outbox = Outbox;
//# sourceMappingURL=outbox.js.map